import contextlib
import doctest
import importlib.metadata
import logging
import tempfile
from os.path import basename
from string import Template as Niks

from django.conf import settings
from django.core.files import File
from django.template import Context, Template
from django.test import TestCase, override_settings
from bs4 import BeautifulSoup
from PIL import Image, ImageDraw

import ocyan.main.templateloaders
import ocyan.main.templatetags.ocyan_main
from ocyan.main.template import OcyanDjangoTemplates

logger = logging.getLogger()

TEST_TEMPLATE = Niks("""
{% load ocyan_thumbnail %}
{% $tag pic "200" slug="$slug" alt="dummy" classes="dummy henk" %}
""")


def load_tests(loader, tests, ignore):  # pylint: disable=W0613
    tests.addTests(doctest.DocTestSuite(ocyan.main.templateloaders))
    tests.addTests(doctest.DocTestSuite(ocyan.main.templatetags.ocyan_main))
    return tests


class OcyanMainTestCase(TestCase):
    params = {
        "NAME": "",
        "DIRS": [],
        "APP_DIRS": [],
        "OPTIONS": {"loaders": ["henk", "jos"]},
    }

    def test_cached_template_loaders_non_debug(self):
        # In non-debug we should use django.template.loaders.cached.Loader
        self.params["OPTIONS"]["debug"] = False
        backend = OcyanDjangoTemplates(params=self.params)
        self.assertEqual(
            backend.engine.loaders,
            [("django.template.loaders.cached.Loader", ["henk", "jos"])],
        )

    def test_cached_template_loaders_debug(self):
        # In debug we should *not* use django.template.loaders.cached.Loader
        self.params["OPTIONS"]["debug"] = True
        backend = OcyanDjangoTemplates(params=self.params)
        self.assertEqual(backend.engine.loaders, ["henk", "jos"])

    def test_use_ocyan_backend(self):
        """Make sure we use *our* backend, so we'll get caching"""
        for definition in settings.TEMPLATES:
            if not definition["BACKEND"].endswith("DjangoTemplates"):
                continue
            self.assertEqual(
                definition["BACKEND"], "ocyan.main.template.OcyanDjangoTemplates"
            )

    def test_distribution_metadata(self):
        self.assertEqual(importlib.metadata.version("ocyan.main"), "2.0.1")

    def test_namespace_uses_stdlib_implementation(self):
        import ocyan

        self.assertNotIn("pkg_resources", ocyan.__dict__)


@contextlib.contextmanager
def dummy_image(text, static=False):
    with tempfile.NamedTemporaryFile(
        dir=settings.TEST_STATIC_SOURCE if static else settings.MEDIA_ROOT,
        suffix=".png",
    ) as thefile:
        img = Image.new("RGB", (800, 400), color=(73, 109, 137))

        d = ImageDraw.Draw(img)
        d.text((10, 10), text, fill=(255, 255, 0))

        img.save(thefile, "PNG")
        thefile.seek(0)
        yield File(thefile.file, thefile.name)


@override_settings(
    THUMBNAIL_DEBUG=True,
    CACHES={"default": {"BACKEND": "django.core.cache.backends.dummy.DummyCache"}},
)
class OcyanThumbnailTestCase(TestCase):
    maxDiff = None

    def assert_picture_markup(
        self, result, *, lazy=False, owl=False, width="200", height="100"
    ):
        soup = BeautifulSoup(result, "html.parser")
        picture = soup.find("picture")
        self.assertIsNotNone(picture)

        source_name = "data-src" if lazy else "source"
        source_attribute = "data-srcset" if owl else "srcset"
        sources = picture.find_all(source_name)
        self.assertEqual(len(sources), 2)
        self.assertEqual(sources[0].get("type"), "image/webp")
        for source in sources:
            srcset = source.get(source_attribute, "")
            self.assertTrue(srcset.startswith("/media/cache/"))
            self.assertEqual(srcset.count(","), 2)

        image_name = "data-img" if lazy else "img"
        image_attribute = "data-src" if owl else "src"
        image = picture.find(image_name)
        self.assertIsNotNone(image)
        self.assertTrue(image.get(image_attribute, "").startswith("/media/cache/"))
        self.assertEqual(image.get("width"), width)
        self.assertEqual(image.get("height"), height)
        self.assertEqual(image.get("alt"), "dummy")
        self.assertTrue({"dummy", "henk"}.issubset(set(image.get("class", []))))

        if lazy:
            self.assertIn("lazy", picture.get("class", []))
        if owl:
            self.assertIn("owl-lazy", image.get("class", []))

    def test_picture(self):
        with dummy_image("picture") as pic:
            tpl = Template(TEST_TEMPLATE.substitute(tag="picture", slug="picture"))
            ctx = Context({"pic": pic})
            result = tpl.render(ctx)
            pic.close()

        self.assert_picture_markup(result)

    def test_lazy_picture(self):
        with dummy_image("lazy_picture") as pic:
            tpl = Template(
                TEST_TEMPLATE.substitute(tag="lazy_picture", slug="lazy_picture")
            )
            ctx = Context({"pic": pic})
            result = tpl.render(ctx)

        self.assert_picture_markup(result, lazy=True)

    def test_responsive_image(self):
        with dummy_image("responsive_image") as pic:
            tpl = Template(
                """{% load ocyan_thumbnail %}{% responsive_image pic slug="responsive_image" alt="dummy" classes="dummy henk" %}"""
            )
            ctx = Context({"pic": pic})
            result = tpl.render(ctx)

        self.assert_picture_markup(result, width="800", height="400")

    def test_lazy_responsive_image(self):
        with dummy_image("lazy_responsive_image") as pic:
            tpl = Template(
                """{% load ocyan_thumbnail %}{% lazy_responsive_image pic slug="lazy_responsive_image" alt="dummy" classes="dummy henk" %}"""
            )
            ctx = Context({"pic": pic})
            result = tpl.render(ctx)

        self.assert_picture_markup(result, lazy=True, width="800", height="400")

    def test_lazy_owl_picture(self):
        with dummy_image("lazy_owl_picture") as pic:
            tpl = Template(
                TEST_TEMPLATE.substitute(
                    tag="lazy_owl_picture", slug="lazy_owl_picture"
                )
            )
            ctx = Context({"pic": pic})
            result = tpl.render(ctx)

        self.assert_picture_markup(result, owl=True)

    def test_static_picture(self):
        with dummy_image("static_picture", static=True) as pic:
            tpl = Template(
                TEST_TEMPLATE.substitute(tag="static_picture", slug="static_picture")
            )
            ctx = Context({"pic": basename(pic.name)})
            result = tpl.render(ctx)

        self.assert_picture_markup(result)

    def test_static_lazy_picture(self):
        with dummy_image("static_lazy_picture", static=True) as pic:
            tpl = Template(
                TEST_TEMPLATE.substitute(
                    tag="static_lazy_picture", slug="static_lazy_picture"
                )
            )
            ctx = Context({"pic": basename(pic.name)})
            result = tpl.render(ctx)

        self.assert_picture_markup(result, lazy=True)



# Generated by Django 3.2.20 on 2023-10-13 08:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('main', '0002_alter_cronlock_pid'),
    ]

    operations = [
        migrations.CreateModel(
            name='Maintenance',
            fields=[
                ('key', models.CharField(max_length=255, primary_key=True, serialize=False, unique=True)),
                ('result', models.TextField()),
            ],
        ),
    ]



# Generated by Django 3.2.23 on 2024-01-22 13:03

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('main', '0003_maintenance'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='maintenance',
            options={'ordering': ('pk',)},
        ),
    ]



"""
Clean up all <font> and style= and things that mess up the styling from
rich text fields.
"""

from django.core.management import call_command
from django.core.management.base import BaseCommand

from oscar.core.loading import get_model

from ocyan.main.templatetags.ocyan_main import stripformatting
from ocyan.main.utils import disable_searchengine_updates

Product = get_model("catalogue", "Product")
ProductAttribute = get_model("catalogue", "ProductAttribute")
ProductAttributeValue = get_model("catalogue", "ProductAttributeValue")
Category = get_model("catalogue", "Category")


class Command(BaseCommand):
    help = __doc__

    def handle(self, *args, **options):
        with disable_searchengine_updates():
            for p in Product.objects.all():
                p.description = stripformatting(p.description)
                p.long_description = (
                    stripformatting(p.long_description)
                    if p.long_description is not None
                    else None
                )
                p.save()
            for pa in ProductAttributeValue.objects.filter(
                attribute__type=ProductAttribute.RICHTEXT
            ):
                pa.value_richtext = (
                    stripformatting(pa.value_richtext)
                    if pa.value_richtext is not None
                    else None
                )
                pa.save()
            for c in Category.objects.all():
                c.description = stripformatting(c.description)
                c.save()

        call_command("update_oscar_index")



"""
Your escape from migration hell.

Do not use this. You are being very careless if you do.

Force Run migration without InconsistentMigrationHistory nonsence.

This command will skip checking for consistent migration history and
it will als skip checking for conflicts.

It will just run the command for the app you specified. Sometimes that
is just what you need to get out of a bad situation.

This can also go wrong really fast though! You might mess up you migration
history even worse!
"""

from unittest import mock

from django.core.management.base import BaseCommand

try:
    from ocyan.i18n.db.migrations.loader import I18nMigrationLoader as MigrationLoader
    from ocyan.i18n.management.commands.i18nmigrate import Command as MigrateCommand

    migration_executor_module = "ocyan.i18n.db.migrations.executor.I18nMigrationLoader"
except ImportError:
    from django.core.management.commands.migrate import Command as MigrateCommand
    from django.db.migrations.loader import MigrationLoader

    migration_executor_module = "django.db.migrations.executor.MigrationLoader"


class NoWorriesnMigrationLoader(MigrationLoader):
    "Not a care in the world"

    def check_consistent_history(self, connection):
        pass

    def detect_conflicts(self):
        return []


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument(
            "app_label",
            help="App label of an application to synchronize the state, required for careless migrations",
        )
        parser.add_argument(
            "migration_name",
            nargs="?",
            help="Database state will be brought to the state after that migration.",
        )
        parser.add_argument(
            "--no-dry-run", action="store_true", help="Really commit to the database"
        )

    def handle(self, *args, **options):
        cmd = [
            "manage.py",
            "migrate",
            "--skip-checks",
            "--traceback",
            "--no-input",
            "--no-color",
            options["app_label"],
            options["migration_name"],
        ]
        if not options["no_dry_run"]:
            self.stdout.write(
                self.style.ERROR(
                    "\nAre you sure you want to seriously mess up your migrations table? Then check the help section for this command first.\n\n"
                )
            )
            self.print_help("manage.py", "careless_migrate")
            exit(1)

        with mock.patch(
            migration_executor_module,
            NoWorriesnMigrationLoader,
        ):
            migrate = MigrateCommand(
                stdout=self.stdout, stderr=self.stderr, no_color=True
            )
            migrate.run_from_argv(cmd)



# -*- coding: utf-8 -*-
import logging
from urllib.parse import urljoin

from bs4 import BeautifulSoup
from django import template
from django.contrib.admin.options import IS_POPUP_VAR
from django.contrib.sites.shortcuts import get_current_site
from django.template import TemplateSyntaxError
from django.template.loader_tags import do_include
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from django.utils.safestring import mark_safe

from ocyan.main.loading import get_ocyan_class
from ocyan.main.utils import resolve_entities

register = template.Library()
logger = logging.getLogger(__name__)

NonceNode = get_ocyan_class("main.utils", "NonceNode", module_prefix="ocyan")


@register.simple_tag(takes_context=True)
def is_popup(context):
    request = context.get("request")
    if not request:
        return False
    return IS_POPUP_VAR in request.POST or IS_POPUP_VAR in request.GET


@register.filter
def stripformatting(html):
    """
    Strip formatting attributes from the passed html

    Strips style attributes, font tags style tags and script tags and empty tags.
    Also corrects invalid html

    >>> html = '<p><span style="color: rgb(69, 69, 69); font-family: &quot;Open Sans&quot;, sans-serif; font-size: 13px; font-style: normal; font-variant-ligatures: normal; font-variant-caps: normal; font-weight: 400; letter-spacing: normal; orphans: 2; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 2; word-spacing: 0px; -webkit-text-stroke-width: 0px; background-color: rgb(255, 255, 255); text-decoration-style: initial; text-decoration-color: initial; display: inline !important; float: none;">Ik heet klaas</span></p><p>hai</p>'
    >>> stripformatting(html)
    '<p><span>Ik heet klaas</span></p><p>hai</p>'
    >>> html = '<p><font face="verdana" color="green">This is some text!</font></p><p><font size="3" color="red">This is some text!</font></p>'
    >>> stripformatting(html)
    '<p>This is some text!</p><p>This is some text!</p>'
    >>> html = '<p><style>h1 { 5px; }</style><h1>hai</p>'
    >>> stripformatting(html)
    '<p></p><h1>hai</h1>'
    >>> html = '<p>henk is een dork<br/>en een sukkel</p>'
    >>> stripformatting(html)
    '<p>henk is een dork<br/>en een sukkel</p>'
    >>> html = '<p>henk is een dork<br/>en een sukkel'
    >>> stripformatting(html)
    '<p>henk is een dork<br/>en een sukkel</p>'
    >>> html = "ḷꝉɑ𝒏ммȃṋẵᶑℹҝκḙ"
    >>> stripformatting(html)
    'ḷꝉɑ𝒏ммȃṋẵᶑℹҝκḙ'
    >>> html = '<script type="text/javascript">window.botcoinbot.start()</script><p>hai ik ben onschuldig</p>'
    >>> stripformatting(html)
    '<p>hai ik ben onschuldig</p>'
    >>> html = '<p>I am &apos;&CloseCurlyDoubleQuote;&apos; > &apos;&#32;&apos; > &apos;&#x20;&apos; &gt; and I rock</p>'
    >>> stripformatting(html)
    "<p>I am '”' &gt; ' ' &gt; ' ' &gt; and I rock</p>"
    >>> html = ""
    >>> stripformatting(html)
    ''
    >>> html = None
    >>> stripformatting(html)
    ''
    >>> html = '<iframe width="490" height="344" src="https://www.youtube.com/embed/tiDp2bkj8VQ" frameborder="0" allowfullscreen></iframe>'
    >>> stripformatting(html)
    '<iframe allowfullscreen="" frameborder="0" height="344" src="https://www.youtube.com/embed/tiDp2bkj8VQ" width="490"></iframe>'
    """
    try:
        unescaped_html = resolve_entities(html).strip()
        if not unescaped_html:
            return ""

        # Skip html/body as we don't want them.
        # lxml will add them if they don't already exist.
        soup = BeautifulSoup(unescaped_html, "lxml").html.body
        for junk in soup.find_all(["style", "script"]):
            junk.decompose()

        tags = soup.find_all()
        for tag in tags:
            if tag.name in ["iframe"]:
                continue
            if tag.name == "font":
                tag.replaceWithChildren()
            if not tag.isSelfClosing and len(tag.get_text(strip=True)) == 0:
                tag.decompose()
            else:
                tag.attrs.clear()

        return mark_safe("".join(map(str, soup.contents)))
    except Exception as e:  # pylint: disable=broad-except
        # catch any exception and log so we can improve the filter.
        logger.exception(e)
        return html


@register.simple_tag(takes_context=True)
def absolute_url(context, path_or_name):
    """
    Tries to turns a path into an absolute url.

    Usage::
    >>> from django.template import RequestContext, Template, Context
    >>> from django.test import RequestFactory
    >>> from argparse import Namespace
    >>> factory = RequestFactory()
    >>> request = factory.get('/')
    >>> tpl = Template("{% load ocyan_main %}{% absolute_url path %}")
    >>> tpl.render(RequestContext(request, {"path": "/kanarie/pietje"}))
    'http://testserver/kanarie/pietje'
    >>> tpl.render(RequestContext(request, {"path": "kanarie/pietje"}))
    'http://testserver/kanarie/pietje'
    >>> tpl = Template("{% load ocyan_main %}{% absolute_url path %}")
    >>> tpl.render(Context({"path": "/example/com"}))
    'https://example.com/example/com'
    >>> tpl.render(Context({"path": "example/com"}))
    'https://example.com/example/com'
    >>> site = Namespace(domain="lol.org")
    >>> tpl = Template("{% load ocyan_main %}{% absolute_url path %}")
    >>> tpl.render(Context({"path": "/ru/?exa=true", "site": site}))
    'https://lol.org/ru/?exa=true'
    >>> tpl.render(Context({"path": "ru/?exa=true", "site": site}))
    'https://lol.org/ru/?exa=true'
    >>> tpl.render(Context({"path": "admin:index"}))
    'https://example.com/manage/'
    >>> tpl.render(Context({"path": "admin:index", "site": site}))
    'https://lol.org/manage/'
    >>> tpl.render(RequestContext(request, {"path": "admin:index"}))
    'http://testserver/manage/'
    """
    if path_or_name.startswith("http"):
        return path_or_name
    elif not path_or_name.startswith("/"):
        try:
            path = reverse(path_or_name)
        except NoReverseMatch:
            path = path_or_name
    else:
        path = path_or_name

    if "request" in context:
        request = context["request"]
        return request.build_absolute_uri(path)

    site = context.get("site", get_current_site(context.get("request")))
    return urljoin("https://%s" % site.domain, path)


@register.simple_tag(name="eval", takes_context=True)
def ocyan_eval(context, html):
    try:
        return template.Template(html).render(context)
    except TemplateSyntaxError:
        return mark_safe(html)


def do_include_only(parser, token):
    if "only" not in token.contents:
        raise TemplateSyntaxError(
            "esi_fragment MUST have isolated context, please add 'only' like the include tag: https://docs.djangoproject.com/en/3.2/ref/templates/builtins/#include"
        )
    return do_include(parser, token)


register.tag(name="esi_fragment")(do_include_only)


@register.tag
def nonce(parser, token):  # pylint: disable=unused-argument
    return NonceNode()



"""
Ensure that only 1 process can be running a managment command at the same
time, even across multiple application servers.

Uses a database lock on a table row, in a separate database connection.
"""

import logging
from datetime import timedelta

from django import db
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction
from django.utils import timezone
from django.utils.module_loading import import_string

logger = logging.getLogger(__name__)


class CronLockManager(models.Manager):
    def lock(self, name, pid, max_age=timedelta(hours=12)):
        with transaction.atomic():
            try:
                tha_lock = (
                    CronLock.objects.filter(jobname=name)
                    .select_for_update(nowait=True)
                    .get()
                )

                if tha_lock.pid is None or tha_lock.is_expired(max_age):
                    tha_lock.pid = pid
                    tha_lock.started = timezone.now()
                    tha_lock.save()
                    return tha_lock
            except ObjectDoesNotExist:
                CronLock.objects.create(jobname=name)
                return self.lock(name, pid, max_age)
            except db.DatabaseError:
                logger.debug("Could not obtain lock for %s %s", name, pid)

        return None


class CronLock(models.Model):
    objects = CronLockManager()
    jobname = models.CharField(max_length=255, unique=True)
    started = models.DateTimeField(blank=True, null=True)
    pid = models.PositiveBigIntegerField(blank=True, null=True)

    def is_expired(self, max_age):
        return self.started < (timezone.now() - max_age)

    def __str__(self):
        return self.jobname

    def release(self):
        self.__class__.objects.filter(jobname=self.jobname, pid=self.pid).update(
            pid=None
        )


class MaintenanceManager(models.Manager):
    def open_maintenance(self, jobs):
        completed = self.in_bulk(jobs)
        return [Maintenance(key=job) for job in jobs if job not in completed]


class Maintenance(models.Model):
    objects = MaintenanceManager()
    key = models.CharField(max_length=255, unique=True, primary_key=True)
    result = models.TextField()

    def __str__(self):
        return self.key

    def run(self):
        func = import_string(self.key)
        try:
            result = func()
        except Exception as e:  # pylint: disable=broad-exception-caught
            result = None
            logger.exception(e)

        if result is not None:
            logger.info("Maintenance %s completed: %s", self.key, result)
            self.result = str(result)
            self.full_clean()
            self.save()
        else:
            logger.info(
                "No recordable result for maintenance %s. Maintenance not completed.",
                self.key,
            )

    class Meta:
        ordering = ("pk",)



"""
Drop a table that isn't supposed to be there from the database
"""

from django.core.management.base import BaseCommand
from django.db import connection, transaction


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument("tablename", help="The name of the table to drop")
        parser.add_argument(
            "--no-dry-run", action="store_true", help="Really commit to the database"
        )

    def handle(self, *args, **options):
        with transaction.atomic():
            with connection.cursor() as cursor:
                cursor.execute("DROP TABLE %(tablename)s;" % options)
                if not options["no_dry_run"]:
                    self.stdout.write("Rolling back because --no-dry-run wasn't set")
                    transaction.set_rollback(True)
                else:
                    self.stdout.write("Done")



"""
Use pretranslate to add translations that allready exist to a pofile
"""

import polib
from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils.translation import gettext, override


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument("path", help="path to the pofile you want to check")
        parser.add_argument("-l", "--language", default=settings.LANGUAGE_CODE)
        parser.add_argument("--save", action="store_true")

    def handle(self, *args, **options):
        po = polib.pofile(options["path"])

        for entry in po:
            if not entry.msgstr:
                with override(options["language"]):
                    translation = gettext(entry.msgid)
                    if translation != entry.msgid:
                        entry.msgstr = translation

        if options["save"]:
            po.save(options["path"])
        else:
            self.stdout.write(str(po))



"""
WSGI config for main project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/2.0/howto/deployment/wsgi/
"""

import importlib
import io
import logging
import os

from django.core.wsgi import get_wsgi_application

logger = logging.getLogger()

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "ocyan.main.settings")

application = get_wsgi_application()

# load spooler modules.
from ocyan.core.fender import config  # noqa: E402
from ocyan.core.plugins import get_spooler_mdodules  # noqa: E402

for spooler_module in get_spooler_mdodules():
    importlib.import_module(spooler_module)


from django.conf import settings  # noqa: E402

if not settings.DEBUG:
    # https://uwsgi-docs.readthedocs.io/en/latest/articles/TheArtOfGracefulReloading.html#dealing-with-ultra-lazy-apps-like-django
    # This issues a request to load the Django
    try:
        domain = config.get("django", "domain", "localhost")
        response = application(
            environ={
                "REQUEST_METHOD": "get",
                "REQUEST_URI": "/~startup/",
                "PATH_INFO": "/~startup/",
                "QUERY_STRING": "",
                "wsgi.input": io.StringIO(""),
                "SERVER_NAME": domain,
                "SERVER_PORT": 8000,
            },
            start_response=lambda x, y: None,
        )
        response.close()
    except Exception as exc:  # pylint: disable=W0703
        # Log this in case something went wrong here because of the fake request.
        logger.exception(exc)



import os
import tarfile
import tempfile
from os.path import isdir, isfile, join

from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand

from ocyan.core.fender import config


class Command(BaseCommand):
    help = __doc__
    project = config.get("django", "name", "myproject")

    def add_arguments(self, parser):
        parser.add_argument("path", help="path to the tarbal")
        parser.add_argument(
            "--skip-db",
            action="store_true",
            default=False,
            help="Do not export the database",
        )
        parser.add_argument(
            "--skip-media",
            action="store_true",
            default=False,
            help="Do not export the media folder",
        )

    def handle(self, *args, **options):
        path = options["path"]
        if isfile(path):
            tarname = path
        elif isdir(path):
            tarname = join(path, "%s.tar.gz" % self.project)
        elif path.endswith("tar.gz"):
            tarname = path
        else:
            tarname = "%s.tar.gz" % path

        with tarfile.open(tarname, "w:gz") as tar:
            if options["skip_db"]:
                self.stdout.write(self.style.WARNING("skipping database"))
            else:
                os.umask(0)
                with tempfile.TemporaryDirectory() as tmp:
                    export_dir = join(tmp, self.project)
                    os.chmod(tmp, 0o777)
                    os.mkdir(export_dir, 0o777)
                    db_dump_path = join(export_dir, "db.json")

                    if config.is_webshop:
                        call_command(
                            "dumpdata",
                            "--indent=4",
                            "--exclude=catalogue.ProductCategoryHierarchy",
                            "--output=%s" % db_dump_path,
                        )
                    else:
                        call_command(
                            "dumpdata",
                            "--indent=4",
                            "--all",
                            "--output=%s" % db_dump_path,
                        )

                    tar.add(export_dir, arcname=self.project)

            if options["skip_media"]:
                self.stdout.write(self.style.WARNING("skipping media"))
            else:
                tar.add(settings.MEDIA_ROOT, arcname=join(self.project, "media"))
                tar.add(settings.STATIC_ROOT, arcname=join(self.project, "static"))



"""
Use this command to view the INSTALLED_APPS segments, displaying how the app
overrides are structured.
"""

import json

from django.core.management.base import BaseCommand
from django.core.serializers.json import DjangoJSONEncoder

from ocyan.main.settings.base import _installed_apps


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument(
            "segment",
            choices=("all", "front", "django", "ocyan", "apps", "oscar", "wagtail"),
            default="all",
            help="The segment to display",
            nargs="?",
        )

    def handle(self, *args, **options):
        if options["segment"] != "all":
            segment = getattr(_installed_apps, options["segment"])
            self.stdout.write(
                json.dumps(segment.representation, indent=2, cls=DjangoJSONEncoder)
            )
        else:
            self.stdout.write(str(_installed_apps))



"""
Check if there are apps that override something and have missing migrations.
"""

import importlib
import os
import os.path
from copy import deepcopy
from shutil import copy

from django.apps.registry import Apps
from django.conf import settings
from django.core.management.commands.makemigrations import Command as MakeMigrations
from django.db.migrations.exceptions import BadMigrationError
from django.db.migrations.loader import MIGRATIONS_MODULE_NAME

from ocyan.core.utils import SegmentedInstalledApps, unique_everseen


def get_segmented_installed_apps():
    settings_module = importlib.import_module(settings.SETTINGS_MODULE)
    for _, value in vars(settings_module).items():
        if isinstance(value, SegmentedInstalledApps):
            SEGMENTED_INSTALLED_APPS = value
            break
    else:
        from ocyan.main.settings.base import _installed_apps as SEGMENTED_INSTALLED_APPS

    return SEGMENTED_INSTALLED_APPS


def get_migrations(app_module_name, skip_replaced=True):
    module_name = "%s.%s" % (app_module_name, MIGRATIONS_MODULE_NAME)
    try:
        module = importlib.import_module(module_name)
    except ImportError:
        return set()
    else:
        directory = os.path.dirname(module.__file__)
        migration_names = set()
        for name in os.listdir(directory):
            if name.endswith(".py"):
                import_name = name.rsplit(".", 1)[0]
                if import_name[0] not in "_.~":
                    migration_names.add(import_name)

        # remove all replaced migrations
        if skip_replaced:
            for migration_name in deepcopy(migration_names):
                migration_module = importlib.import_module(
                    "%s.%s" % (module_name, migration_name)
                )
                if not hasattr(migration_module, "Migration"):
                    raise BadMigrationError(
                        "Migration %s in app %s has no Migration class"
                        % (migration_name, app_module_name)
                    )
                if hasattr(migration_module.Migration, "replaces"):
                    _migration_names = [
                        _migration_name
                        for app_label, _migration_name in migration_module.Migration.replaces
                        if app_module_name.endswith(app_label)
                    ]
                    migration_names -= set(_migration_names)

        return migration_names


INSTALLED_APPS = get_segmented_installed_apps()


class Command(MakeMigrations):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument(
            "--dry-run",
            action="store_true",
            dest="dry_run",
            help="Just show what migrations would be made; don't actually write them.",
        )
        parser.add_argument(
            "--include-replaced",
            action="store_true",
            dest="include_replaced",
            help="Also copy replaced migrations, usually not needed",
        )
        parser.add_argument(
            "segment",
            choices=("front", "django", "ocyan", "apps", "oscar", "wagtail"),
            help="Which segment would you like to check?",
        )
        parser.add_argument(
            "--no-header",
            action="store_false",
            dest="include_header",
            help="Do not add header comments to new migration file(s).",
        )

    def handle(self, *_, **options):
        segment = options.pop("segment")
        self.segment = getattr(  # pylint: disable=attribute-defined-outside-init
            INSTALLED_APPS, segment
        )
        self.skip_replaced = not options.pop(  # pylint: disable=attribute-defined-outside-init
            "include_replaced"
        )
        self.apps = Apps(  # pylint: disable=attribute-defined-outside-init
            unique_everseen(self.segment)
        )
        app_labels = self.apps.app_configs.keys()

        configuration = dict(
            options,
            interactive=True,
            merge=False,
            empty=False,
            name=None,
            check_changes=False,
        )
        super().handle(*app_labels, **configuration)
        self.stdout.write(self.style.HTTP_REDIRECT("KLAAR DUS"))

    # pylint: disable=unused-argument
    def write_migration_files(self, changes, update_previous_migration_paths=None):
        if bool(changes):
            self.stdout.write(
                self.style.ERROR(
                    "Missing migrations detected, checking ancestor apps for missing migrations"
                )
            )

            for app_label, _ in changes.items():
                app_config = self.apps.app_configs[app_label]
                app_migrations = get_migrations(app_config.name)
                for ancestor_app in self.segment.extensions_from_app_root(
                    app_config.name
                ):
                    ancestor_migrations = get_migrations(
                        ancestor_app, skip_replaced=self.skip_replaced
                    )
                    missing_migrations = ancestor_migrations.difference(app_migrations)

                    # find first app that has migrations not present in the current app.
                    if missing_migrations:
                        self.stdout.write(
                            self.style.SQL_TABLE(
                                "Missing migrations for %s found in %s."
                                % (app_config.name, ancestor_app)
                            )
                        )
                        for migration in missing_migrations:
                            # copy the migration to the app that needs it
                            migration_module_name = "%s.%s.%s" % (
                                ancestor_app,
                                MIGRATIONS_MODULE_NAME,
                                migration,
                            )
                            self.stdout.write(
                                self.style.HTTP_SUCCESS(
                                    "Copying %s to migration directory of %s"
                                    % (migration, app_config.name)
                                )
                            )
                            try:
                                migration_module = importlib.import_module(
                                    migration_module_name
                                )
                            except ImportError:
                                continue
                            else:
                                if not self.dry_run:
                                    migration_file = migration_module.__file__
                                    target_path = os.path.join(
                                        app_config.path, MIGRATIONS_MODULE_NAME
                                    )
                                    if not os.path.isdir(target_path):
                                        os.mkdir(target_path)
                                    copy(migration_file, target_path)
                        break  # do maximum of 1 app at a time.
        else:
            self.stdout.write(
                self.style.SUCCESS("No changes found everything is up to date.")
            )



"""
Use this command to output some settings
"""

import json
import operator

from django.conf import settings as all_settings
from django.core.management.base import BaseCommand
from django.core.serializers.json import DjangoJSONEncoder


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument("setting", nargs="+")

    def handle(self, *args, **options):
        selected_settings = options["setting"]
        get_settings = operator.attrgetter(*selected_settings)
        settings = get_settings(all_settings)
        if len(selected_settings) == 1:
            out = json.dumps(settings, indent=4, cls=DjangoJSONEncoder)
        else:
            out = json.dumps(
                {k: v for k, v in zip(options["setting"], settings)},
                indent=4,
                cls=DjangoJSONEncoder,
            )
        self.stdout.write(out)



"""
Run maintenance jobs registered with Entrypoint.contribute_to_maintenance

A job is a dotted modules path to a function that takes no argument and
will mark the maintenance job as completed when finishing without errors.
Jobs results are collected in Maintenance models.
Jobs that do not return results will run every deployment.
"""

from django.core.management.base import BaseCommand
from django.db import OperationalError, ProgrammingError

from ocyan.core.plugins import get_maintenances

from ocyan.main.models import Maintenance


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument("--list", action="store_true", help="Show maintenance jobs")
        parser.add_argument(
            "--force",
            help="force rerunning an allready ran job, requires maintenance job key, try --list",
            action="append",
        )

    def handle(self, *args, **options):
        jobs = get_maintenances()
        if options["list"]:
            open_maintenances = [
                henk.key for henk in Maintenance.objects.open_maintenance(jobs)
            ]
            for job in jobs:
                if job in open_maintenances:
                    self.stdout.write("%s: open" % job)
                else:
                    self.stdout.write("%s: completed" % job)

        elif options["force"]:
            for job in options["force"]:
                try:
                    Maintenance.objects.get(key=job).run()
                except (OperationalError, ProgrammingError, Maintenance.DoesNotExist):
                    Maintenance(key=job).run()
        else:
            for job in Maintenance.objects.open_maintenance(jobs):
                job.run()



from django.conf import settings
from django.template.backends.base import BaseEngine
from django.template.backends.django import DjangoTemplates
from django.template.engine import Engine


class OcyanEngine(Engine):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if (
            not self.debug
            and self.loaders[0][0] != "django.template.loaders.cached.Loader"
        ):
            self.loaders = [("django.template.loaders.cached.Loader", self.loaders)]


class OcyanDjangoTemplates(DjangoTemplates):
    def __init__(self, params):  # pylint: disable=W0231
        # Copied over from parent class
        params = params.copy()
        options = params.pop("OPTIONS").copy()
        options.setdefault("autoescape", True)
        options.setdefault("debug", settings.DEBUG)
        options.setdefault(
            "file_charset", getattr(settings, "FILE_CHARSET", settings.DEFAULT_CHARSET)
        )
        libraries = options.get("libraries", {})
        options["libraries"] = self.get_templatetag_libraries(libraries)
        # calls parent's super class
        BaseEngine.__init__(self, params)  # pylint: disable=W0233
        # Use our engine
        self.engine = OcyanEngine(self.dirs, self.app_dirs, **options)



# Generated by Django 3.2.19 on 2023-08-15 08:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('main', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cronlock',
            name='pid',
            field=models.PositiveBigIntegerField(blank=True, null=True),
        ),
    ]



"""
Fix the catalogue migrations
"""

from django.core.management.base import BaseCommand
from django.db import connection, transaction
from django.db.migrations import recorder

from ocyan.core.fender import config

STANDARD = [
    "0001_initial",
    "0002_auto_20150217_1221",
    "0003_data_migration_slugs",
    "0004_auto_20150217_1710",
    "0005_auto_20150604_1450",
    "0006_auto_20150807_1725",
    "0007_auto_20151207_1440",
    "0008_auto_20160304_1652",
    "0009_slugfield_noop",
    "0010_auto_20170420_0439",
    "0011_auto_20170422_1355",
    "0012_auto_20170609_1902",
    "0013_auto_20170821_1548",
    "0014_auto_20181115_1953",
    "0015_product_is_public",
    "0016_auto_20190327_0757",
    "0017_category_is_public",
    "0018_ocyan_additions",
]

TYPED = ["0019_auto_20190711_1230"]

PAID = ["0020_product_options_through", "0021_auto_20190711_1244"]


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument("--no-dry-run", action="store_true")

    def migration_names(self):
        # pylint: disable=no-member
        if "ocyan.plugin.oscar_paid_product_options" in config.ocyan_plugins:
            return STANDARD + TYPED + PAID
        elif "ocyan.plugin.oscar_typed_product_options" in config.ocyan_plugins:
            return STANDARD + TYPED
        return STANDARD

    def handle(self, *args, **options):
        rec = recorder.MigrationRecorder(connection)
        rec.ensure_schema()
        is_up_to_date = True
        for name in STANDARD:
            is_up_to_date = (
                is_up_to_date
                and rec.migration_qs.filter(app="catalogue").filter(name=name).exists()
            )

        if is_up_to_date:
            self.stdout.write(
                self.style.SUCCESS(
                    "No need to mess with migration history, it is all good"
                )
            )
        else:
            with transaction.atomic():
                self.stdout.write(
                    self.style.WARNING(
                        "Migrations need fixing, deleting all migrations"
                    )
                )
                rec.migration_qs.filter(app="catalogue").delete()
                self.stdout.write(self.style.WARNING("Writing new migrations"))
                for name in self.migration_names():
                    rec.migration_qs.create(app="catalogue", name=name)
                if not options[
                    "no_dry_run"
                ]:  # rollback everything because it waS DRY RUN LOL
                    self.stdout.write(
                        self.style.ERROR("Rolling back import because of dry run.")
                    )
                    transaction.set_rollback(True)



import sys

from django.core.management import BaseCommand
from PIL import Image
from PIL.PngImagePlugin import PngImageFile

from oscar.apps.catalogue.models import MissingProductImage
from oscar.core.loading import get_model

ProductImage = get_model("catalogue", "ProductImage")
Category = get_model("catalogue", "Category")


def set_matte(image, matte):
    try:
        if not isinstance(image, MissingProductImage):
            pilimage = Image.open(image)
            if isinstance(pilimage, PngImageFile):
                new_pic = Image.new("RGB", pilimage.size, matte)
                rgba_png = pilimage.convert("RGBA")
                new_pic.paste(rgba_png, rgba_png)
                image.close()
                with image.open("wb") as target_image:
                    new_pic.save(target_image)
                new_pic.close()

            pilimage.close()
            image.close()
    except OSError as e:
        sys.stderr.write("Broken image found %s, %s\n" % (image, e))


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--matte", default="#ffffff", help="Color as a hexadecimal string."
        )

    def handle(self, *args, **options):
        hexstr = options["matte"].strip("#")
        r, g, b = int(hexstr[:2], 16), int(hexstr[2:4], 16), int(hexstr[4:], 16)

        for image in ProductImage.objects.all():
            set_matte(image.original, (r, g, b))

        for category in Category.objects.all():
            if category.image:
                set_matte(category.image, (r, g, b))



from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import QuerySet

from oscar.core.loading import get_model, is_model_registered


class OcyanJSONEncoder(DjangoJSONEncoder):
    """
    Encoding support for various objects that are used in ocyan projects.
    """

    def default(self, o):
        if hasattr(o, "to_json"):
            return o.to_json()

        if hasattr(o, "json"):
            return o.json()

        if isinstance(o, QuerySet):
            return list(o)

        if is_model_registered("catalogue", "AttributeOption"):
            AttributeOption = get_model("catalogue", "AttributeOption")
            if isinstance(o, AttributeOption):
                return o.option

        return super().default(o)



from pathlib import Path
from tempfile import gettempdir

BASE_DIR = Path(gettempdir()) / "ocyan-main-tests"
TEST_STATIC_SOURCE = BASE_DIR / "static-source"
for directory in (BASE_DIR / "static", TEST_STATIC_SOURCE, BASE_DIR / "media"):
    directory.mkdir(parents=True, exist_ok=True)

SECRET_KEY = "ocyan-main-test-key"
DEBUG = True
ALLOWED_HOSTS = ["testserver", "localhost"]

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.staticfiles",
    "compressor",
    "ocyan.main",
    "sorl.thumbnail",
]

DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}}
DEFAULT_AUTO_FIELD = "django.db.models.AutoField"
SITE_ID = 1

ROOT_URLCONF = "ocyan.main.test_urls"
STATIC_URL = "/static/"
STATIC_ROOT = BASE_DIR / "static"
STATICFILES_DIRS = [TEST_STATIC_SOURCE]
MEDIA_URL = "/media/"
MEDIA_ROOT = BASE_DIR / "media"

TEMPLATES = [
    {
        "BACKEND": "ocyan.main.template.OcyanDjangoTemplates",
        "DIRS": [],
        "APP_DIRS": False,
        "OPTIONS": {
            "context_processors": ["django.template.context_processors.request"],
            "loaders": ["django.template.loaders.app_directories.Loader"],
        },
    }
]

FORM_RENDERER = "ocyan.main.renderers.OcyanFormRenderer"
CACHES = {"default": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache"}}
USE_TZ = True



# Generated by Django 3.2.19 on 2023-07-10 12:49

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CronLock',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('jobname', models.CharField(max_length=255, unique=True)),
                ('started', models.DateTimeField(blank=True, null=True)),
                ('pid', models.PositiveIntegerField(blank=True, null=True)),
            ],
        ),
    ]



# pylint: disable=unused-argument
from django.contrib import admin

from .models import CronLock, Maintenance


class CronLockAdmin(admin.ModelAdmin):
    readonly_fields = ("jobname", "started", "pid")

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False


class MaintenanceAdmin(admin.ModelAdmin):
    readonly_fields = ("key", "result")

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False


admin.site.register(CronLock, CronLockAdmin)
admin.site.register(Maintenance, MaintenanceAdmin)



from django.http import HttpResponse
from django.urls import include, path


def admin_index(request):
    return HttpResponse()


admin_patterns = ([path("", admin_index, name="index")], "admin")
urlpatterns = [path("manage/", include(admin_patterns, namespace="admin"))]



from django.http import Http404
from django.shortcuts import render


def serve_engine_page(request, slug):
    """
    Dedicated route for Template Engine pages.
    Avoids root-level slug collisions with catalogue/detail routes.
    """
    try:
        from ocyan.plugin.template_engine.engine.models.pages import EnginePage
    except Exception as exc:
        raise Http404("Template Engine is unavailable") from exc

    page = EnginePage.objects.live().public().filter(slug=slug).first()
    if page is None:
        raise Http404("Engine page not found")

    return page.specific.serve(request)


def startup(request):
    return render(request, "ocyan/startup.html")



"""main URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""

from django.conf import settings
from django.conf.urls.i18n import i18n_patterns
from django.conf.urls.static import static
from django.contrib import admin
from django.urls import path, re_path, reverse_lazy
from django.views.generic.base import RedirectView

from ocyan.core.fender import config
from ocyan.core.plugins import get_plugins

from .views import serve_engine_page, startup

plugins = get_plugins()

# the default
handler404 = settings.OCYAN_404_HANDLER

urlpatterns = [
    path(r"~startup/", startup),
    path(r"pages/<slug:slug>/", serve_engine_page, name="engine-page-by-slug"),
    path(r"manage/", admin.site.urls),
    re_path(r"^django_admin/.*", RedirectView.as_view(url=reverse_lazy("admin:index"))),
    re_path(r"^django-admin/.*", RedirectView.as_view(url=reverse_lazy("admin:index"))),
]
if config.i18n_enabled:
    urlpatterns = i18n_patterns(*urlpatterns, prefix_default_language=False)

urlpatterns = plugins.pipe("contribute_to_urlpatterns", urlpatterns)

urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)



"""
Delete duplicate product images
"""

from collections import defaultdict

from django.core.management.base import BaseCommand

from oscar.core.loading import get_model

from oscarapi.utils.files import file_hash

Product = get_model("catalogue", "Product")


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument(
            "--no-dry-run", action="store_true", help="Really commit to the database"
        )

    def handle(self, *args, **options):
        for p in Product.objects.filter(images__isnull=False):
            pic_table = defaultdict(list)
            for im in p.images.all():
                p = file_hash(im.original)
                pic_table[p].append(im)

            for imagelist in pic_table.values():
                if len(imagelist) > 1:
                    _, *duplicates = imagelist
                    for duplicate in duplicates:
                        duplicate.delete()



"""
Destroy the database and replace with a dump from a tar.gz file. Also import media files
"""

import os
import subprocess
import tarfile
import tempfile
from os.path import exists, join

from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.db import connection

from ocyan.core.fender import config

from ocyan.main import utils
from ocyan.main.signals import pre_import_destructive


class Command(BaseCommand):
    help = __doc__
    project = config.get("django", "name", "myproject")

    def add_arguments(self, parser):
        parser.add_argument("path", help="path to the tarbal")
        parser.add_argument(
            "--skip-db",
            action="store_true",
            default=False,
            help="Do not import the database",
        )
        parser.add_argument(
            "--skip-media",
            action="store_true",
            default=False,
            help="Do not import the media folder",
        )
        parser.add_argument(
            "--command",
            help="command to use for importing, default is loaddata",
            default="loaddata",
        )
        parser.add_argument(
            "--ignorenonexistent",
            "-i",
            action="store_true",
            dest="ignore",
            help="Ignores entries in the serialized data for fields that do not currently exist on the model.",
        )
        parser.add_argument(
            "--disable-constraint-checking",
            help="Disable constraint checking while importing, which helps when importing a broken database",
            action="store_true",
        )
        parser.add_argument(
            "--anon",
            action="store_true",
            help="anoymise user data using the destroy_userata command",
        )

    def handle(self, *args, **options):
        os.umask(0)
        with tempfile.TemporaryDirectory() as tmp:
            os.chmod(tmp, 0o777)
            with tarfile.open(options["path"]) as tar:
                tar.extractall(tmp)

                if options["skip_media"]:
                    self.stdout.write(self.style.WARNING("skipping media"))
                else:
                    if not exists(settings.MEDIA_ROOT):
                        os.mkdir(settings.MEDIA_ROOT)
                    subprocess.check_call(
                        [
                            "/usr/bin/rsync",
                            "--archive",
                            "--verbose",
                            join(tmp, self.project, "media", ""),
                            join(settings.MEDIA_ROOT, ""),
                        ]
                    )

                    if not exists(settings.STATIC_ROOT):
                        os.mkdir(settings.STATIC_ROOT)
                    subprocess.check_call(
                        [
                            "/usr/bin/rsync",
                            "--archive",
                            "--verbose",
                            join(tmp, self.project, "static", ""),
                            join(settings.STATIC_ROOT, ""),
                        ]
                    )

                if options["skip_db"]:
                    self.stdout.write(self.style.WARNING("skipping database"))
                else:
                    db_dump_path = join(tmp, self.project, "db.json")

                    with utils.disable_searchengine_updates():
                        call_command("flushall")
                        if options["disable_constraint_checking"]:
                            connection.disable_constraint_checking()
                        pre_import_destructive.send(sender=self.__class__)
                        command_args = [db_dump_path]
                        if options["ignore"]:
                            command_args.append("--ignorenonexistent")
                        call_command(options["command"], *command_args)
                        if options["disable_constraint_checking"]:
                            connection.enable_constraint_checking()

                        call_command("migrate", "--fake")

        if options["anon"]:
            call_command("destroy_userdata")

        self.stdout.write(self.style.SUCCESS("done."))



import functools
import os.path

from django.apps import apps
from django.apps.config import AppConfig
from django.template.loaders import filesystem

from ocyan.core.utils import unique_everseen

from ocyan.main.settings.base import _installed_apps as INSTALLED_APPS


@functools.lru_cache()
def get_app_extension_template_dirs(dirname):
    """
    Return an iterable of paths of directories to load app templates from.
    This version also adds any replaced apps and app extensions, in the
    correct order of priority.
    dirname is the name of the subdirectory containing templates inside
    installed applications.

    >>> slash = lambda x: "/".join(str(x).rsplit("/", maxsplit=3)[1:])
    >>> fuck_up_some_slashes = lambda templates: set([slash(x) for x in templates])

    >>> from django.template.utils import get_app_template_dirs

    >>> templates = get_app_template_dirs('templates')
    >>> old_app_loader_templates = fuck_up_some_slashes(templates)

    >>> templates2 = get_app_extension_template_dirs('templates')
    >>> new_app_loader_templates = fuck_up_some_slashes(templates2)

    >>> old_app_loader_templates.issubset(new_app_loader_templates)
    True
    """
    template_dirs = []

    app_name_table = {app_cfg.name: app_cfg for app_cfg in apps.get_app_configs()}
    for app_config in apps.get_app_configs():
        section = INSTALLED_APPS.section_for_app_extension(app_config.name)
        if section is not None:
            for app_ext in section.extensions_from_app_root(app_config.name):
                if app_ext in app_name_table:
                    ext_app_config = app_name_table[app_ext]
                else:
                    ext_app_config = AppConfig.create(app_ext)

                if not ext_app_config.path:
                    continue
                template_dir = os.path.join(ext_app_config.path, dirname)
                if os.path.isdir(template_dir):
                    template_dirs.append(template_dir)
        else:
            if not app_config.path:
                continue
            template_dir = os.path.join(app_config.path, dirname)
            if os.path.isdir(template_dir):
                template_dirs.append(template_dir)

    # Immutable return value because it will be cached and shared by callers.
    return tuple(unique_everseen(template_dirs))


class OcyanExtendedAppTemplateLoader(filesystem.Loader):
    def get_dirs(self):
        return get_app_extension_template_dirs("templates")



"""
Use this command to delete every single thing in the database
"""

from django.core.management.base import CommandError
from django.core.management.color import no_style
from django.core.management.commands import flush
from django.core.management.sql import sql_flush
from django.db import connections


class Command(flush.Command):
    help = __doc__

    def handle(self, **options):
        interactive = options["interactive"]
        database = options["database"]
        connection = connections[database]

        self.style = no_style()

        sql_list = sql_flush(
            self.style,
            connection,
            reset_sequences=True,
            allow_cascade=True,
        )

        if interactive:
            confirm = input("""You have requested a flush of the database.
This will IRREVERSIBLY DESTROY all data currently in the %r database,
and return each table to an empty state.
Are you sure you want to do this?

    Type 'yes' to continue, or 'no' to cancel: """ % connection.settings_dict["NAME"])
        else:
            confirm = "yes"

        if confirm == "yes":
            try:
                connection.ops.execute_sql_flush(sql_list)
            except Exception as exc:
                raise CommandError(
                    "Database %s couldn't be flushed. Possible reasons:\n"
                    "  * The database isn't running or isn't configured correctly.\n"
                    "  * At least one of the expected database tables doesn't exist.\n"
                    "  * The SQL was invalid.\n"
                    "Hint: Look at the output of 'django-admin sqlflush'. "
                    "That's the SQL this command wasn't able to run.\n"
                    % (connection.settings_dict["NAME"],)
                ) from exc
        else:
            self.stdout.write("Flush cancelled.\n")



from pathlib import Path

from setuptools import find_packages, setup

PACKAGE_CLASSIFIERS = [
    "Framework :: Ocyan",
]

install_requires: list = [
    "ocyan.core>=1.2.7",
    "django>=5,<6",
    "django-oscar",
    "django-oscar-api",
    "django-compressor",
    "django-libsass",
    "beautifulsoup4",
    "lxml",
    "polib",
    "sorl-thumbnail",
]

extras_require: dict = {
    "test": [
        "wheel",
        "empty_testproject",
        "pylint-django",
        "isort",
        "ruff",
        "pillow",
        "ocyan.plugin.testing",
        "vdt.versionplugin.wheel",
        "coverage",
        "ocyan.plugin.oscar_paid_product_options",
        "ocyan.plugin.seo",
        "oxyan.themes",
    ]
}

setup(
    name="ocyan.main",
    version="2.0.1",
    url="https://git.mandelblog.com/mandel-plugins/ocyan.main",
    author="Mandel",
    author_email="info@mandelblog.com",
    description="Integrates and configures ocyan plugins in projects.",
    long_description=Path("README.rst").read_text(encoding="utf-8"),
    long_description_content_type="text/x-rst",
    include_package_data=False,
    packages=find_packages(include=["ocyan", "ocyan.*"]),
    package_data={"ocyan.main": ["templates/**/*", "locale/**/*"]},
    python_requires=">=3.12,<3.13",
    install_requires=install_requires,
    classifiers=PACKAGE_CLASSIFIERS,
    extras_require=extras_require,
)



"""
Anonymize all userdata present in the database
"""

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand

from oscar.core.loading import get_model

User = get_user_model()

UserAddress = get_model("address", "UserAddress")
CommunicationEvent = get_model("order", "CommunicationEvent")
ShippingAddress = get_model("order", "ShippingAddress")
BillingAddress = get_model("order", "BillingAddress")
ProductReview = get_model("reviews", "ProductReview")

try:
    Subscription = get_model("newsletter", "Subscription")
except LookupError:
    Subscription = ShippingAddress


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument(
            "--no-dry-run", action="store_true", help="Really commit to the database"
        )

    def handle(self, *args, **options):
        User.objects.all().update(email="", first_name="", last_name="")
        ProductReview.objects.all().update(email="", name="Anonymous")
        UserAddress.objects.all().delete()
        CommunicationEvent.objects.all().delete()
        ShippingAddress.objects.all().delete()
        BillingAddress.objects.all().delete()
        Subscription.objects.all().delete()



"""
Django settings for main project.

Generated by 'django-admin startproject' using Django 2.0.6.

For more information on this file, see
https://docs.djangoproject.com/en/2.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.0/ref/settings/
"""

from ocyan.core.fender import config
from ocyan.core.plugins import (
    add_plugin_settings,
    get_installed_apps,
    get_middleware,
    get_sass_functions,
)

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
X_FRAME_OPTIONS = "SAMEORIGIN"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ["localhost", "127.0.0.1", "[::1]"]
domain = config.get("django", "domain", "")
if domain:
    ALLOWED_HOSTS.append(domain)

AUTHENTICATION_BACKENDS = ["django.contrib.auth.backends.ModelBackend"]

COMPRESS_ENABLED = True
COMPRESS_PRECOMPILERS = (("text/x-scss", "django_libsass.SassCompiler"),)
COMPRESS_CSS_FILTERS = [
    "compressor.filters.css_default.CssAbsoluteFilter",
    "compressor.filters.cssmin.rCSSMinFilter",
]

LOGIN_REDIRECT_URL = "/"

LOGGING = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "verbose": {
            "format": "%(name)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s"
        }
    },
    "root": {"level": "INFO", "handlers": ["console"]},
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "verbose",
        },
        "null": {"class": "logging.NullHandler"},
    },
    "loggers": {
        "ocyan": {"level": "DEBUG", "propagate": True},
        "configtype": {"level": "DEBUG"},
        "django": {"level": "INFO", "propagate": True},
        "django.request": {"level": "ERROR", "propagate": True},
        "oscar.alerts": {"level": "ERROR", "propagate": True},
        # "elasticsearch": {"level": "DEBUG", "propagate": True},
        "elastic_transport": {"level": "ERROR", "propagate": True},
        "uwsgidecorators": {"level": "ERROR", "propagate": True},
        # "django.db.backends": {"level": "DEBUG", "propagate": True},
        "django.template": {"level": "WARNING", "propagate": True},
        "sorl.thumbnail": {"level": "CRITICAL", "propagate": True},
    },
}

# Application definition

_installed_apps = get_installed_apps(
    django=[
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
        "django.contrib.sites",
        "ocyan.main",
    ],
    apps=["compressor"],
)

INSTALLED_APPS = _installed_apps.unique_list()

LIBSASS_CUSTOM_FUNCTIONS = get_sass_functions()

MIDDLEWARE = get_middleware(
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
)

ROOT_URLCONF = "ocyan.main.urls"

TEMPLATES = [
    {
        "BACKEND": "ocyan.main.template.OcyanDjangoTemplates",
        "DIRS": [],
        "NAME": "django",
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.static",
                "django.template.context_processors.media",
                "django.template.context_processors.i18n",
                "django.contrib.messages.context_processors.messages",
            ],
            "loaders": [
                "django.template.loaders.filesystem.Loader",
                # use instead of django.template.loaders.app_directories.Loader
                # to enforce searching for templates in replaced apps and extended apps
                "ocyan.main.templateloaders.OcyanExtendedAppTemplateLoader",
            ],
        },
    }
]

FORM_RENDERER = "ocyan.main.renderers.OcyanFormRenderer"

WSGI_APPLICATION = "ocyan.main.wsgi.application"

# Password validation
# https://docs.djangoproject.com/en/2.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]


# Internationalization
# https://docs.djangoproject.com/en/2.0/topics/i18n/

LANGUAGE_CODE = config.get("django", "language_code", "") or "nl"

# make the 404 handler changeable in settings
OCYAN_404_HANDLER = "django.views.defaults.page_not_found"

TIME_ZONE = "UTC"
USE_I18N = True
USE_L10N = True
USE_TZ = True

# The default email backend is one that doesn't break.
EMAIL_BACKEND = "django.core.mail.backends.console.EmailBackend"

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.0/howto/static-files/

MEDIA_URL = "/media/"
STATIC_URL = "/static/"

STATICFILES_FINDERS = [
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    "compressor.finders.CompressorFinder",
]

# https://docs.djangoproject.com/en/5.1/ref/settings/#storages
STORAGES = {
    "default": {
        "BACKEND": "django.core.files.storage.FileSystemStorage",
    },
    "staticfiles": {
        "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage",
    },
}

SITE_ID = 1

FILE_UPLOAD_PERMISSIONS = 0o644

COMPRESS_CSS_HASHING_METHOD = None

WAGTAILIMAGES_WEBP_QUALITY = WAGTAILIMAGES_JPEG_QUALITY = THUMBNAIL_QUALITY = 60
THUMBNAIL_PROGRESSIVE = False

# 14days * seconds per day (86400)
OCYAN_RANDOM_TOKEN_TIMEOUT = 14 * 86400

# Import settings from Ocyan plugins
add_plugin_settings(locals())

# make other changes here



import logging

from django.apps import apps

from oscar.core.exceptions import (
    AppNotFoundError,
)

# loading settings via django.conf triggers registry exceptions but this
# is safe because these settings are just defaults for all ocyan projects.
# the actual settings are still constructed in the project itself.
from oscar.core.exceptions import (
    ModuleNotFoundError as OscarModuleNotFoundError,
)
from oscar.core.loading import (  # pylint: disable=unused-import
    _import_module,
    _pluck_classes,
    get_model,  # noqa: F401 - public compatibility re-export
    is_model_registered,  # noqa: F401 - public compatibility re-export
)

from ocyan.main.settings.base import _installed_apps as INSTALLED_APPS

logger = logging.getLogger(__name__)


try:
    from oscar.core.loading import _find_installed_apps_entry
except ImportError:  # in oscar 3 this function has been removed

    def _find_installed_apps_entry(module_label):
        """
        Given a module label, finds the name of the matching Oscar app from the
        Django app registry.
        """
        if "dashboard" in module_label:
            split_module_label = module_label.split(".")[:2]
            potential_label = "_".join(list(reversed(split_module_label)))
            if potential_label.endswith("_dashboard"):
                try:
                    app_config = apps.get_app_config(potential_label)
                    app_label = ".".join(split_module_label)
                    return app_config.name, app_label
                except LookupError:
                    pass

        app_label = module_label.split(".")[0]

        try:
            app_config = apps.get_app_config(app_label)
        except LookupError:
            raise AppNotFoundError(
                "Couldn't find an app to import %s from" % module_label
            )

        return app_config.name, app_label


class OcyanClassLoader(object):
    def get_extension_module_names(self, oscar_app_name, installed_apps_entry):
        """
        Retrieve the names of the extension modules that should be searched for classes.
        """
        INSTALLED_APPS_SECTION = INSTALLED_APPS.section_for_app_extension(
            oscar_app_name
        )
        if INSTALLED_APPS_SECTION is None:
            INSTALLED_APPS_SECTION = INSTALLED_APPS.section_for_app_extension(
                installed_apps_entry
            )
            if (
                INSTALLED_APPS_SECTION is None
            ):  # nog steeds? Return dan maar de gevraagde app, er zijn geen overrides.
                return [installed_apps_entry]

        if oscar_app_name == installed_apps_entry:
            return INSTALLED_APPS_SECTION.extensions_for_app(oscar_app_name)
        else:
            extensions = INSTALLED_APPS_SECTION.extensions_for_app(oscar_app_name)
            # regular app overrides don't register extensions,
            # register it here if it is missing.
            # the reason that we want it registered, is that the ExtendableInstalledAppsItem
            # is layered. it could be possible that extensions is not empty at this
            # point but that it is just missing the entire registration tree behind
            # installed_apps_entry. That is why we register it here, to make sure
            # the layers that are allready present are not discarded entirely in
            # favor of installed_apps_entry's layers
            if installed_apps_entry not in extensions:
                INSTALLED_APPS_SECTION.register_extension(
                    installed_apps_entry, oscar_app_name
                )
                return INSTALLED_APPS_SECTION.extensions_for_app(oscar_app_name)
            return extensions

    def get_extension_modules(
        self, module_label, classnames, app_name, oscar_app_name, installed_apps_entry
    ):
        """
        Get all the extension modules that should be searched for classes.
        """
        extensions = self.get_extension_module_names(
            oscar_app_name, installed_apps_entry
        )
        extension_modules = []
        for application_extension in extensions:
            sub_module = module_label.replace(app_name, "", 1)
            application_extension_module_label = "%s%s" % (
                application_extension,
                sub_module,
            )
            application_extension_module = _import_module(
                application_extension_module_label, classnames
            )
            extension_modules.append(application_extension_module)

        return extension_modules

    def oscar_class_loader_part(self, module_label, classnames, module_prefix):
        ###############################################################
        # The following code had to be COPY PASTED FROM OSCAR
        # COPIED FROM oscar.core.loading.default_class_loader
        ###############################################################

        if "." not in module_label:
            raise ValueError("Importing from top-level modules is not supported")

        oscar_module_label = "%s.%s" % (module_prefix, module_label)
        oscar_module = _import_module(oscar_module_label, classnames)

        installed_apps_entry, app_name = _find_installed_apps_entry(module_label)
        ###############################################################
        # COPY PASTED CODE STOPS HERE
        ###############################################################
        return oscar_module, installed_apps_entry, app_name

    def __call__(self, module_label, classnames, module_prefix):
        # do the same as oscar
        oscar_module, installed_apps_entry, app_name = self.oscar_class_loader_part(
            module_label, classnames, module_prefix
        )

        oscar_app_name = "%s.%s" % (module_prefix, app_name)
        extension_modules = self.get_extension_modules(
            module_label, classnames, app_name, oscar_app_name, installed_apps_entry
        )
        extension_modules.append(oscar_module)

        if not any(extension_modules):
            raise OscarModuleNotFoundError(
                "The module with label '%s' could not be imported. This either "
                "means that it indeed does not exist, or you might have a problem "
                "with a circular import." % module_label
            )

        return _pluck_classes(extension_modules, classnames)


class OcyanBaseClassLoader(OcyanClassLoader):
    def __init__(self, extension_root):
        self.extension_root = extension_root

    def get_extension_module_names(self, oscar_app_name, installed_apps_entry):
        extensions = super().get_extension_module_names(
            oscar_app_name, installed_apps_entry
        )
        try:
            index = extensions.index(self.extension_root)
            return extensions[index + 1 :]
        except ValueError:
            raise OscarModuleNotFoundError(
                "trying to load baseclass not derivative of extension_root"
            )


ocyan_class_loader = OcyanClassLoader()


def _get_baseclasses(module_label, classnames, extension_root, module_prefix):
    loader = OcyanBaseClassLoader(extension_root)
    return loader(module_label, classnames, module_prefix)


def get_baseclass(module_label, classname, extension_root, module_prefix="oscar.apps"):
    return _get_baseclasses(module_label, [classname], extension_root, module_prefix)[0]


def get_baseclasses(
    module_label, classnames, extension_root, module_prefix="oscar.apps"
):
    return _get_baseclasses(module_label, classnames, extension_root, module_prefix)


def get_ocyan_baseclass(
    module_label, classname, extension_root, module_prefix="ocyan.plugin"
):
    return _get_baseclasses(module_label, [classname], extension_root, module_prefix)[0]


def get_ocyan_baseclasses(
    module_label, classnames, extension_root, module_prefix="ocyan.plugin"
):
    return _get_baseclasses(module_label, classnames, extension_root, module_prefix)


def get_ocyan_class(module_label, classname, module_prefix="ocyan.plugin"):
    return get_ocyan_classes(module_label, [classname], module_prefix)[0]


def get_ocyan_classes(module_label, classnames, module_prefix="ocyan.plugin"):
    return ocyan_class_loader(module_label, classnames, module_prefix)



import re
import threading
from contextlib import contextmanager
from copy import deepcopy
from datetime import datetime, time, timedelta
from html import entities

from django.conf import settings
from django.contrib.auth.tokens import PasswordResetTokenGenerator
from django.core.management import call_command
from django.template.defaulttags import CsrfTokenNode
from django.test.utils import override_settings
from django.utils.crypto import constant_time_compare
from django.utils.http import base36_to_int

from .models import CronLock, logger

ENTITY_RE = re.compile(r"&(\w+?;)")

# create search engine settings, with autoupdate turned OFF
if hasattr(settings, "WAGTAILSEARCH_BACKENDS"):
    NO_AUTOUPDATE_SEARCH = deepcopy(settings.WAGTAILSEARCH_BACKENDS)
    NO_AUTOUPDATE_SEARCH["default"]["AUTO_UPDATE"] = False
else:
    NO_AUTOUPDATE_SEARCH = {}


def disable_searchengine_updates():
    SETTINGS_MODULE = settings.SETTINGS_MODULE
    return override_settings(
        WAGTAILSEARCH_BACKENDS=NO_AUTOUPDATE_SEARCH, SETTINGS_MODULE=SETTINGS_MODULE
    )


def _unicode_for_entity_with_name(name):
    # if the entity is unknown, insert niks
    return entities.html5.get(name, " ")


def escape_single(matchobj):
    return _unicode_for_entity_with_name(matchobj.group(1))


def resolve_entities(string):
    """Resolve all html entities to their corresponding unicode character"""
    if not string:
        return ""

    return ENTITY_RE.sub(escape_single, string)


class RandomTokenGenerator(PasswordResetTokenGenerator):
    key_salt = "ocyan.plugin.oscar_checkout.checkout.utils.RandomTokenGenerator"

    def check_token(self, user, token):
        """
        Check that a password reset token is correct for a given user.
        """
        if not (user and token):
            return False
        # Parse the token
        try:
            ts_b36, _ = token.split("-")
            # Support pre-Django 3.1 generated tokens.
            legacy_token = len(ts_b36) < 4
        except ValueError:
            return False

        try:
            ts = base36_to_int(ts_b36)
        except ValueError:
            return False

        # Check that the timestamp/uid has not been tampered with
        if not constant_time_compare(
            self._make_token_with_timestamp(user, ts, self.secret), token
        ):
            # Support pre-Django 3.1 generated tokens.
            if not constant_time_compare(
                self._make_token_with_timestamp(user, ts, self.secret),
                token,
            ):
                return False

        # convert days to seconds and round to
        # midnight (server time) for pre-Django 3.1 generated tokens.
        now = self._now()
        if legacy_token:
            ts *= 24 * 60 * 60
            ts += int((now - datetime.combine(now.date(), time.min)).total_seconds())

        # Check the timestamp is within limit.
        if (self._num_seconds(now) - ts) > settings.OCYAN_RANDOM_TOKEN_TIMEOUT:
            return False

        return True


randomtokengenerator = RandomTokenGenerator()


@contextmanager
def acquire_cronlock(name, max_age=timedelta(hours=12)):
    thread_id = threading.get_ident()
    try:
        lock = CronLock.objects.lock(name, thread_id, max_age)
        yield bool(lock)
    finally:
        if lock is not None:
            lock.release()


def run_exclusive(name, *args, max_age=timedelta(hours=12), **kwargs):
    """
    Run a management command inside a database lock.

    This ensures that only 1 process can be running the command at the same
    time, even across multiple application servers.
    """
    thread_id = threading.get_ident()
    lock = CronLock.objects.lock(name, thread_id, max_age)

    if lock is not None:
        try:
            logger.debug("Running command %s", name)
            call_command(name, *args, **kwargs)
        finally:
            lock.release()


class NonceNode(CsrfTokenNode):
    "Templatetag node implementation for the nonce templatetag"

    def render(self, context):
        nonce = context.get("csrf_token")
        if nonce:
            return "%s==" % nonce

        return super().render(context)



from os import makedirs
from os.path import basename, dirname, join
from posixpath import normpath
from shutil import copyfile

from django import template
from django.conf import settings
from django.contrib.staticfiles import finders
from django.core.files.storage import default_storage
from sorl.thumbnail import default
from sorl.thumbnail.conf import settings as thumbnail_settings
from sorl.thumbnail.parsers import parse_geometry

try:
    from wagtail.images.models import Filter
    from wagtail.images.shortcuts import get_rendition_or_not_found

    WAGTAIL_THUMBNAILS_AVAILABLE = True
except Exception:  # pragma: no cover
    Filter = None
    get_rendition_or_not_found = None
    WAGTAIL_THUMBNAILS_AVAILABLE = False

register = template.Library()

MAXIMUM_SUPPORTED_SCREEN_WIDTH = 5120


def to_webp(image, options=None):
    if image is None:
        return None

    # pylint: disable=protected-access
    return default.backend.get_thumbnail(
        image,
        str(int(image.width)),
        upscale=False,
        format="WEBP",
        **(options or {}),
    )


class FakeFile(object):
    "fake django.core.files.File object"

    def __init__(self, name):
        self.name = name


def static_to_media(path):
    """
    Copy a static file to the media folder, if not allready present, so
    thumbnails can be generated.

    This only checks by filename, so if you update your static images the
    media representations will not be updated at all!
    """
    if not default_storage.exists(path):
        normalized_path = normpath(path).lstrip("/")
        absolute_path = finders.find(normalized_path)
        media_dir = join(settings.MEDIA_ROOT, dirname(path))
        makedirs(media_dir, mode=0o755, exist_ok=True)
        target_path = join(media_dir, basename(path))
        copyfile(absolute_path, target_path)

    return path


# pylint: disable=redefined-builtin


@register.inclusion_tag("ocyan/thumbnail/picture.html", takes_context=True)
def picture(
    context,
    image,
    geometry,
    format=thumbnail_settings.THUMBNAIL_FORMAT,
    alt=None,
    upscale=False,
    classes="img-fluid",
    **options,
):
    width, height = parse_geometry(geometry)
    return {
        "width": width,
        "height": height,
        "image": image,
        "geometry": geometry,
        "format": format,
        "alt": alt if alt is not None else context.get("alt"),
        "upscale": upscale,
        "classes": classes,
        "options": options,
    }


@register.inclusion_tag("ocyan/thumbnail/lazy_picture.html", takes_context=True)
def lazy_picture(
    context,
    image,
    geometry,
    format=thumbnail_settings.THUMBNAIL_FORMAT,
    alt=None,
    upscale=None,
    classes="img-fluid",
    **options,
):
    pic = picture(
        context,
        image,
        geometry,
        format=format,
        alt=alt,
        upscale=upscale,
        classes=classes,
        **options,
    )
    pic["lazy"] = True
    return pic


@register.inclusion_tag("ocyan/thumbnail/responsive_image.html", takes_context=True)
def responsive_image(
    context,
    image,
    format=thumbnail_settings.THUMBNAIL_FORMAT,
    alt=None,
    classes="img-fluid",
    mobile=None,
    **options,
):
    """
    responsive_image will create 3 different resolution images based on the size of the uploaded image.
    the ratios are: 100% 50% and 30%

    If needed a separate image can be passed for mobile devices. This is useful for large hero graphics
    that are landscape on desktop but must be in portrait mode for mobile.

    Example for wagtail image, this will crap the image around the focal point::
        {% load wagtailimages_tags ocyan_thumbnail %}

        {% image self.image fill-200x400-c100 as mobile %}
        {% lazy_responsive_image  self.image.file mobile=mobile.file %}

    Example for oscar image::
        {% load ocyan_thumbnail sorl_thumbnail %}

        {% thumbnail image '200x400' as portrait %}
        {% lazy_responsive_image  image mobile=portrait %}
        {% endthumbnail %}

    Note that you don't have to scale the image, you can also supply a completely
    different image.
    """
    pic = picture(
        context,
        image,
        str(MAXIMUM_SUPPORTED_SCREEN_WIDTH),
        format=format,
        alt=alt,
        upscale=False,
        classes=classes,
        mobile=mobile,
        **options,
    )
    pic["mobile"] = mobile
    pic["mobile_webp"] = to_webp(mobile)

    return pic


@register.inclusion_tag(
    "ocyan/thumbnail/lazy_responsive_image.html", takes_context=True
)
def lazy_responsive_image(
    context,
    image,
    format=thumbnail_settings.THUMBNAIL_FORMAT,
    alt=None,
    classes="img-fluid",
    mobile=None,
    **options,
):
    pic = lazy_picture(
        context,
        image,
        str(MAXIMUM_SUPPORTED_SCREEN_WIDTH),
        format=format,
        alt=alt,
        upscale=False,
        classes=classes,
        **options,
    )
    pic["mobile"] = mobile
    pic["mobile_webp"] = to_webp(mobile)

    return pic


@register.inclusion_tag("ocyan/thumbnail/owl_lazy_picture.html", takes_context=True)
def lazy_owl_picture(
    context,
    image,
    geometry,
    format=thumbnail_settings.THUMBNAIL_FORMAT,
    alt=None,
    upscale=None,
    classes="img-fluid",
    **options,
):
    return lazy_picture(
        context,
        image,
        geometry,
        format=format,
        alt=alt,
        upscale=upscale,
        classes=classes,
        **options,
    )


@register.inclusion_tag("ocyan/thumbnail/static_picture.html", takes_context=True)
def static_picture(
    context,
    staticfile,
    geometry,
    format=thumbnail_settings.THUMBNAIL_FORMAT,
    alt=None,
    upscale=None,
    classes="img-fluid",
    **options,
):
    "Try not to use this, thnx"
    media_path = static_to_media(staticfile)
    pic = picture(
        context,
        media_path,
        geometry,
        format=format,
        alt=alt,
        upscale=upscale,
        classes=classes,
        **options,
    )
    # pylint: disable=protected-access
    pic["format"] = default.backend._get_format(FakeFile(media_path))
    return pic


@register.inclusion_tag("ocyan/thumbnail/static_lazy_picture.html", takes_context=True)
def static_lazy_picture(
    context,
    staticfile,
    geometry,
    format=thumbnail_settings.THUMBNAIL_FORMAT,
    alt=None,
    upscale=None,
    classes="img-fluid",
    **options,
):
    "Try to never use this, only to quickfix annoying lighthouse issues in customizations"
    pic = static_picture(
        context,
        staticfile,
        geometry,
        format=format,
        alt=alt,
        upscale=upscale,
        classes=classes,
        **options,
    )
    pic["lazy"] = True
    return pic


@register.simple_tag(takes_context=True)
def retina(context, thumb, multiplier, options=None):
    # pylint: disable=protected-access
    try:
        _format = default.backend._get_format(thumb)
    except AttributeError:  # it is the dummy image haha
        _format = "JPEG"

    times = float(multiplier)
    return default.backend.get_thumbnail(
        context["image"],
        "%sx%s" % (int(thumb.width * times), int(thumb.height * times)),
        upscale=False,
        format=_format,
        **(options or {}),
    )


@register.inclusion_tag("ocyan/thumbnail/focuspoint_picture.html")
def focuspoint_picture(
    wagtailimage,
    classes="img-fluid",
):
    """
    This templatetage will render a picture tag with images with sizes 2560, 1200 and 768 pixel wide::

        <picture>
            <source srcset="henk.width-2560.jpg 2560w, henk.width-2560.fill-1200x800-c100.jpg 1200w, henk.width-2560.fill-768x512-c100.jpg 768w">
            <source type="image/webp" srcset="henk.width-2560.format-webp.webp 2560w, henk.width-2560.fill-1200x800-c100.format-webp.webp 1200w, henk.width-2560.fill-768x512-c100.format-webp.webp 768w">
            <img src="henk.width-2560.jpg" alt="alt" width="2560" height="1080" class="img-fluid">
        </picture>

    If the images has a focal area defined, the image will use the focal area and crop
    towards the focal area. If not, the image will just be scaled.

    If the image is too small to get the renditions 2560, 1200 and 768, no upscaling
    will be performed and the images that can not be rendered will be left out.
    """
    if not WAGTAIL_THUMBNAILS_AVAILABLE or wagtailimage is None:
        return {"img": wagtailimage, "classes": classes}

    img2560 = get_rendition_or_not_found(wagtailimage, Filter(spec="width-2560"))
    img2560_webp = get_rendition_or_not_found(
        wagtailimage, Filter(spec="width-2560|format-webp")
    )
    img1200 = None
    img1200_webp = None
    img768 = None
    img768_webp = None

    if wagtailimage.has_focal_point():
        img1200_scale = 1200 / img2560.width
        if img1200_scale < 1.0:
            img1200 = get_rendition_or_not_found(
                wagtailimage,
                Filter(spec="fill-1200x%s-c50" % (int(img2560.height * img1200_scale))),
            )
            img1200_webp = get_rendition_or_not_found(
                wagtailimage,
                Filter(
                    spec="fill-1200x%s-c50|format-webp"
                    % (int(img2560.height * img1200_scale))
                ),
            )

        img768_scale = 768 / img2560.width
        if img768_scale < 1.0:
            img768 = get_rendition_or_not_found(
                wagtailimage,
                Filter(spec="fill-768x%s-c100" % (int(img2560.height * img768_scale))),
            )
            img768_webp = get_rendition_or_not_found(
                wagtailimage,
                Filter(
                    spec="fill-768x%s-c100|format-webp"
                    % (int(img2560.height * img768_scale))
                ),
            )
    else:
        if img2560.width > 1200:
            img1200 = get_rendition_or_not_found(
                wagtailimage, Filter(spec="width-1200")
            )
            img1200_webp = get_rendition_or_not_found(
                wagtailimage, Filter(spec="width-1200|format-webp")
            )

        if img2560.width > 768:
            img768 = get_rendition_or_not_found(wagtailimage, Filter(spec="width-768"))
            img768_webp = get_rendition_or_not_found(
                wagtailimage, Filter(spec="width-768|format-webp")
            )

    return {
        # pass the original image along for the cache key
        "img": wagtailimage,
        "img2560": img2560,
        "img1200": img1200,
        "img768": img768,
        "img2560_webp": img2560_webp,
        "img1200_webp": img1200_webp,
        "img768_webp": img768_webp,
        "classes": classes,
    }


@register.inclusion_tag("ocyan/thumbnail/lazy_focuspoint_picture.html")
def lazy_focuspoint_picture(
    wagtailimage,
    classes="img-fluid",
):
    """
    Same as focuspoint_picture, but lazy.
    """
    return focuspoint_picture(wagtailimage, classes)


@register.inclusion_tag("ocyan/thumbnail/lazy_tns_focuspoint_picture.html")
def lazy_tns_focuspoint_picture(wagtailimage, classes="img-fluid"):
    return focuspoint_picture(wagtailimage, classes)



from django.conf import settings
from django.forms.renderers import DjangoTemplates
from django.utils.functional import cached_property


class OcyanFormRenderer(DjangoTemplates):
    """
    This class is needed because django uses a different template renderer for
    form widgets than for regular templates. We do need the loaders we have
    set for the regular templates to be the same for form templates.
    Otherwise we can not oveeride form widgets in app extensions.
    """

    @cached_property
    def engine(self):
        engine = super().engine  # pylint: disable=E1101

        TEMPLATES = settings.TEMPLATES[0]

        return self.backend(
            {
                "APP_DIRS": False,
                "DIRS": engine.dirs,
                "NAME": "djangoforms",
                "OPTIONS": {"loaders": TEMPLATES["OPTIONS"].get("loaders")},
            }
        )



import json

from django import template
from django.utils.safestring import mark_safe
from django.utils.translation import gettext, gettext_noop

from ocyan.core.fender import config

register = template.Library()

gettext_noop("false")
gettext_noop("true")


@register.simple_tag(name="ocyanjson")
def get_value_from_ocyan_json(section, key, fallback=None):
    return config.get(section, key, fallback)


@register.filter
def parse(value):
    return json.loads(value)


def _printable(value):
    if isinstance(value, bool):
        val = json.dumps(value)
        translated_val = gettext(val)
        return '<span class="printable-bool printable-bool-%s">%s</span>' % (
            str(value).lower(),
            translated_val,
        )
    elif isinstance(value, int):
        return '<span class="printable-int">%s</span>' % value
    elif isinstance(value, float):
        return '<span class="printable-float">%s</span>' % value
    elif value is None:
        return '<span class="printable-empty"/>'
    elif isinstance(value, list):
        return '<span class="printable-flattened-list">%s</span>' % ", ".join(
            map(_printable, value)
        )

    return gettext(value)


@register.filter
def printable(value):
    return mark_safe(_printable(value))
